import os
import subprocess
import json
from email.parser import BytesHeaderParser


PIP_PATH = os.getenv('PIP_PATH', 'pip3')
TIMEOUT = 15


class PipTree():
    @classmethod
    def _generate_console_output(cls):
        """Take the output of the dependency tree and print to console.
        """
        print(f'Generating Pip Tree report for "{PIP_PATH}"...')
        console_output, number_of_dependencies = cls.generate_dependency_tree()
        print(json.dumps(console_output, indent=4))
        print(f'Pip Tree report complete! {number_of_dependencies} dependencies found for "{PIP_PATH}".')  # noqa

    @classmethod
    def generate_dependency_tree(cls):
        """Generate the dependency tree of your pip virtual environment
        and print to console.
        """
        package_list = cls.get_pip_package_list()
        dependency_tree, number_of_dependencies = cls.get_package_dependencies(package_list)
        return dependency_tree, number_of_dependencies

    @classmethod
    def get_pip_package_list(cls):
        """Get the pip package list of the virtual environment.
        """
        try:
            command = f'{PIP_PATH} list --format=json --disable-pip-version-check --isolated --no-input'
            package_list_output = subprocess.check_output(
                command,
                stdin=None,
                stderr=None,
                shell=True,
                timeout=TIMEOUT
            )
        except subprocess.TimeoutExpired:
            raise subprocess.TimeoutExpired(command, TIMEOUT)
        except subprocess.CalledProcessError:
            raise subprocess.CalledProcessError(127, command)
        return json.loads(package_list_output)

    @classmethod
    def get_package_dependencies(cls, package_list):
        """Get a single package dependencies and return a json object
        """
        final_list = []
        package_count = 0
        for package in package_list:
            try:
                command = f'{PIP_PATH} show {package["name"]} --disable-pip-version-check  --isolated --no-input'
                package_output = subprocess.check_output(
                    command,
                    stdin=None,
                    stderr=None,
                    shell=True,
                    timeout=TIMEOUT
                )
            except subprocess.TimeoutExpired:
                raise subprocess.TimeoutExpired(command, TIMEOUT)
            except subprocess.CalledProcessError:
                raise subprocess.CalledProcessError(127, command)
            parsed_package_output = BytesHeaderParser().parsebytes(
                package_output)
            final_package_output = {
                'name': parsed_package_output['Name'],
                'version': parsed_package_output['Version'],
                'requires': parsed_package_output['Requires'],
                'required-by': parsed_package_output['Required-by'],
            }
            final_list.append(final_package_output)
            package_count += 1
        return final_list, package_count


def main():
    PipTree._generate_console_output()


if __name__ == '__main__':
    main()



import mock
import pytest
import subprocess
import json
from pip_tree import PipTree


MOCK_PACKAGE_LIST = [
    {
        "name": "aiohttp",
        "version": "3.6.2",
        "requires": "async-timeout, multidict, attrs, yarl, chardet",
        "required-by": "slackclient"
    },
    {
        "name": "astroid",
        "version": "2.4.2",
        "requires": "six, wrapt, lazy-object-proxy",
        "required-by": ""
    },
    {
        "name": "async-timeout",
        "version": "3.0.1",
        "requires": "",
        "required-by": "aiohttp"
    }
]


@mock.patch('pip_tree.tree.PipTree.generate_dependency_tree', return_value=['mock-output', 40])
def test__generate_console_output(mock_dependency_tree):
    PipTree._generate_console_output()
    mock_dependency_tree.assert_called_once()


@mock.patch('pip_tree.tree.PipTree.get_pip_package_list', return_value=[{'name': 'mock-package'}])
@mock.patch('pip_tree.tree.PipTree.get_package_dependencies', return_value=[{'Name': 'mock-package'}, 40])  # noqa
def test_generate_dependency_tree(mock_get_package_dependencies, mock_get_pip_package_list):
    PipTree.generate_dependency_tree()
    mock_get_pip_package_list.assert_called_once()
    mock_get_package_dependencies.assert_called_once_with([{'name': 'mock-package'}])


@mock.patch('subprocess.check_output', return_value=json.dumps([{'Name': 'mock-package'}]))
def test_get_pip_package_list_works(mock_subprocess):
    output = PipTree.get_pip_package_list()
    assert isinstance(output, list)


@mock.patch('subprocess.check_output', side_effect=subprocess.TimeoutExpired(cmd=subprocess.check_output, timeout=0.1))  # noqa
def test_get_pip_package_list_timeout(mock_timeout):
    with pytest.raises(subprocess.TimeoutExpired):
        PipTree.get_pip_package_list()


@mock.patch('subprocess.check_output', side_effect=subprocess.CalledProcessError(returncode=127, cmd=subprocess.check_output))  # noqa
def test_get_pip_package_list_error(mock_error):
    with pytest.raises(subprocess.CalledProcessError):
        PipTree.get_pip_package_list()


@pytest.mark.skip('Need to mock stdout before this will work')
@mock.patch('subprocess.check_output', return_value='TODO')
def test_get_package_dependencies_works(mock_subprocess):
    final_list, package_count = PipTree.get_package_dependencies(MOCK_PACKAGE_LIST)
    assert isinstance(final_list, dict)


@mock.patch('subprocess.check_output', side_effect=subprocess.TimeoutExpired(cmd=subprocess.check_output, timeout=0.1))  # noqa
def test_get_package_dependencies_timeout(mock_timeout):
    package_list = [{'name': 'mock-package'}]
    with pytest.raises(subprocess.TimeoutExpired):
        PipTree.get_package_dependencies(package_list)


@mock.patch('subprocess.check_output', side_effect=subprocess.CalledProcessError(returncode=127, cmd=subprocess.check_output))  # noqa
def test_get_package_dependencies_error(mock_error):
    package_list = [{'name': 'mock-package'}]
    with pytest.raises(subprocess.CalledProcessError):
        PipTree.get_package_dependencies(package_list)
